@@ -7,12 +7,6 @@ test_that("SingleR works in DE mode", {
     expect_true(sum(diag(tab))/sum(tab) > 0.95)
 })
 
-test_that("SingleR works in SD mode", {
-    out <- SingleR(test=test, ref=training, labels=training$label, genes="sd")
-    tab <- table(out$labels, test$label)
-    expect_true(sum(diag(tab))/sum(tab) > 0.85) # not as good as 'de'.
-})
-
 test_that("SingleR works with custom gene selection", {
     all.labs <- sort(unique(training$label))
     collected <- rep(list(tail(rownames(training), 100)), length(all.labs))
@@ -41,7 +35,7 @@ test_that("SingleR works when genes are not the same between test and training",
 })
 
 library(Matrix)
-test_that("SingleR works with non-ordinary matrices", {
+test_that("SingleR works with sparse matrices", {
     test.s <- test
     training.s <- training
     assay(test.s) <- as(assay(test.s), "dgCMatrix")
@@ -51,30 +45,36 @@ test_that("SingleR works with non-ordinary matrices", {
     ref <- SingleR(test=test, ref=training, labels=training$label)
     expect_identical(out, ref)
 
+    # Also in parallel.
+    ref2 <- SingleR(test=test.s, ref=training.s, labels=training.s$label, num.threads=3)
+    expect_identical(out, ref2)
+
+    # Works with clustering mode.
     out <- SingleR(test=test.s, ref=training.s, labels=training.s$label, clusters=test.s$label)
     ref <- SingleR(test=test, ref=training, labels=training$label, clusters=test$label)
     expect_identical(out, ref)
 })
 
-test_that("SingleR works with multiple references", {
-    out <- SingleR(test, list(training, training), list(training$label, training$label), recompute=FALSE)
-    ref <- SingleR(test, training, training$label)
-    expect_identical(out$scores, cbind(ref$scores, ref$scores))
+library(DelayedArray)
+test_that("SingleR handles DelayedArray inputs", {
+    ref1 <- SingleR(test=test, ref=training, labels=training$label)
 
-    # Handles mismatching rownames.
+    dtest <- DelayedArray(logcounts(test)) * 10
+    dtrain <- DelayedArray(logcounts(training)) * 10
+    ref2 <- SingleR(test=dtest, ref=dtrain, labels=training$label)
+    expect_identical(ref1, ref2)
+
+    # Also in parallel.
+    ref3 <- SingleR(test=dtest, ref=dtrain, labels=training$label, num.threads=3)
+    expect_identical(ref1, ref3)
+})
+
+test_that("SingleR works with multiple references", {
+    # Handles mismatching row names.
     chosen0 <- sample(rownames(training), 900)
     chosen1 <- sample(rownames(training), 900)
     chosen2 <- sample(rownames(training), 900)
 
-    out <- SingleR(test[chosen0,], list(training[chosen1,], training[chosen2,]), 
-        list(training$label, training$label), recompute=FALSE)
-    inter <- Reduce(intersect, list(chosen0, chosen1, chosen2))
-    ref <- SingleR(test[inter,], list(training[inter,], training[inter,]), 
-        list(training$label, training$label), recompute=FALSE)
-
-    out$reference <- ref$reference <- NULL # basically tied anyway.
-    expect_identical(out, ref)
-
     # Works with recomputation.
     out <- SingleR(test[chosen0,], list(training[chosen1,], training[chosen2,]), 
         list(training$label, training$label), recompute=TRUE)
@@ -86,24 +86,6 @@ test_that("SingleR works with multiple references", {
     expect_identical(out, ref)
 })
 
-test_that("SingleR handles changes to the block size", {
-    ref1 <- SingleR(test=test, ref=training, labels=training$label, genes="de")
-    set.seed(10)
-    ref2 <- SingleR(test, list(training, training), list(training$label, training$label), recompute=FALSE)
-
-    oldb <- DelayedArray::getAutoBlockSize()
-    DelayedArray::setAutoBlockSize(100)
-
-    out1 <- SingleR(test=test, ref=training, labels=training$label, genes="de")
-    set.seed(10)
-    out2 <- SingleR(test, list(training, training), list(training$label, training$label), recompute=FALSE)
-
-    expect_identical(ref1, out1)
-    expect_identical(ref2, out2)
-    
-    DelayedArray::setAutoBlockSize(oldb)
-})
-
 test_that("SingleR handles data.frame inputs", {
     set.seed(10)
     ref1 <- SingleR(test=test, ref=training, labels=training$label)
@@ -129,4 +111,3 @@ test_that("SingleR handles NAs in the labels", {
 
     expect_identical(ref1, ref2)
 })
-
