@@ -63,12 +63,46 @@ test_that("heatmap - can pass excess pheatmap::pheatmap parameters through plotS
         5)
 })
 
-test_that("heatmap scores color can be adjusted when 'normalize = FALSE'", {
+test_that("heatmap scores color can be adjusted, regardless of 'normalize' value", {
     expect_equal(
         plotScoreHeatmap(results = pred, silent = TRUE, return.data = TRUE,
             normalize = FALSE,
             color = colorRampPalette(c("red", "blue"))(33))$color,
         colorRampPalette(c("red", "blue"))(33))
+    expect_equal(
+        plotScoreHeatmap(results = pred, silent = TRUE, return.data = TRUE,
+            normalize = TRUE,
+            color = colorRampPalette(c("red", "blue"))(33))$color,
+        colorRampPalette(c("red", "blue"))(33))
+})
+
+test_that("heatmap allows users to adjust breaks, legend_breaks, legend_labels", {
+    expect_s3_class(
+        plotScoreHeatmap(results = pred, silent = TRUE,
+            normalize = FALSE,
+            color = colorRampPalette(c("red", "blue"))(33),
+            breaks = seq(-5, 5, length.out = 34),
+            legend_breaks = c(-5, 0, 5),
+            legend_labels = c("manually", "set", "labels")),
+        "pheatmap")
+    non_norm_args <- plotScoreHeatmap(results = pred, silent = TRUE, return.data = TRUE,
+        normalize = FALSE,
+        color = colorRampPalette(c("red", "blue"))(33),
+        breaks = seq(-5, 5, length.out = 34),
+        legend_breaks = c(-5, 0, 5),
+        legend_labels = c("manually", "set", "labels"))
+    expect_equal(non_norm_args$breaks, seq(-5, 5, length.out = 34))
+    expect_equal(non_norm_args$legend_breaks, c(-5, 0, 5))
+    expect_equal(non_norm_args$legend_labels, c("manually", "set", "labels"))
+    norm_args <- plotScoreHeatmap(results = pred, silent = TRUE, return.data = TRUE,
+        normalize = TRUE,
+        color = colorRampPalette(c("red", "blue"))(33),
+        breaks = seq(-5, 5, length.out = 34),
+        legend_breaks = c(-5, 0, 5),
+        legend_labels = c("manually", "set", "labels"))
+    expect_equal(norm_args$breaks, seq(-5, 5, length.out = 34))
+    expect_equal(norm_args$legend_breaks, c(-5, 0, 5))
+    expect_equal(norm_args$legend_labels, c("manually", "set", "labels"))
 })
 
 test_that("heatmap is adjusted properly when 'labels.use' yields 1 or 0 labels", {
